/*
 * HPT3xx UDMA/ATA RAID Controller Device Driver for Linux
 * Copyright (C) 2000-2003 HighPoint Technologies, Inc.
 * proc.c  /proc file system support
 */
#include "global.h"
#include "hptintf.h"
#include "oslinux.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,3,0)
/*
 * proc entry
 */
struct proc_dir_entry proc_scsi_hpt3xx = { 
	0, 							/* inode number - ignore. filled by proc_register_dynamic */
	sizeof(PROC_DIR_NAME)-1, 	/* length of name */
	PROC_DIR_NAME,              /* file name */
	S_IFDIR|S_IRUGO|S_IXUGO,	/* */
	2							/* number of links */
};
#endif

/*************************************************************************
 * hpt_set_asc_info
 * Usage:
 *   echo "hpt rebuild start" > /proc/scsi/hpt374/0
 *   echo "hpt rebuild stop" > /proc/scsi/hpt374/0
 *   echo "hpt verify start x" > /proc/scsi/hpt374/0
 *   echo "hpt verify stop x" > /proc/scsi/hpt374/0
 *   echo "hpt dbglevel x" > /proc/scsi/hpt374/0
 *        x can be 0, 1, 2, 3
 *************************************************************************/
int hpt_rescan_all(void);
int hpt_add_disk_to_array(DEVICEID idArray, DEVICEID idDisk);

int hpt_set_asc_info(char *buffer,int length)
{
	int orig_length = length+4;

	if (length>=7 && strncmp(buffer, "rescan ", 7)==0) {
		buffer+=7;
		length-=7;
		if (length>=3 && strncmp(buffer, "all", 3)==0) {
			hpt_rescan_all();
			return orig_length;
		} else if (length>=5 && buffer[1]==',' && buffer[3]==',') {
            /*do nothing now*/
			return orig_length;
		}
	}
#if DBG
	else if (length>=9 && strncmp(buffer, "dbglevel ", 9)==0) {
	    buffer+=9;
		length-=9;
		if (length>=1 && *buffer>='0' && *buffer<='3') {
			hpt_dbg_level = *buffer-'0';
			return orig_length;
		}
	}
#endif
	return -EINVAL;
}

int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						);

int hpt_set_info(char *buffer,int length)
{
	int retval;
#ifdef SUPPORT_IOCTL
	PUCHAR ke_area;
	int err;
	DWORD dwRet;
	PHPT_IOCTL_PARAM32 piop=(PHPT_IOCTL_PARAM32)buffer;
#endif
	if (length >= 4) {
		if (strncmp(buffer,"hpt ",4) == 0) {
			return (retval = hpt_set_asc_info(buffer+4, length-4)) >= 0? retval : -EINVAL;
		}

#ifdef SUPPORT_IOCTL
        if (piop->Magic == HPT_IOCTL_MAGIC) {
        	KdPrintE(("ioctl=%d in=%x len=%d out=%x len=%d\n", 
        		piop->dwIoControlCode,
        		piop->lpInBuffer,
        		piop->nInBufferSize,
        		piop->lpOutBuffer,
        		piop->nOutBufferSize));
			/* 
        	 * verify user buffer.
        	 */
        	if (piop->nInBufferSize) {
				if (hpt_verify_area(VERIFY_READ, (void *)(ULONG_PTR)piop->lpInBuffer, piop->nInBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
			if (piop->nOutBufferSize) {
				if (hpt_verify_area(VERIFY_WRITE, (void *)(ULONG_PTR)piop->lpOutBuffer, piop->nOutBufferSize)) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
			if (piop->lpBytesReturned) {
				if (hpt_verify_area(VERIFY_WRITE, (void *)(ULONG_PTR)piop->lpBytesReturned, sizeof(DWORD))) {
					KdPrintE(("Got bad user address.\n"));
					return -EINVAL;
				}
			}
			/*
        	 * map buffer to kernel.
        	 */
        	if (piop->nInBufferSize+piop->nOutBufferSize > 0x10000) {
        		KdPrintE(("User buffer too large\n"));
        		return -EINVAL;
        	}
        	ke_area = kmalloc(piop->nInBufferSize+piop->nOutBufferSize, GFP_ATOMIC);
			if (ke_area == NULL) {
				KdPrintE(("Couldn't allocate kernel mem.\n"));
				return -EINVAL;
			}
			if (piop->nInBufferSize)
				copy_from_user(ke_area, (void *)(ULONG_PTR)piop->lpInBuffer, piop->nInBufferSize);
			/*
			 * call kernel handler.
			 */
			err = Kernel_DeviceIoControl(piop->dwIoControlCode, ke_area, piop->nInBufferSize,
					ke_area + piop->nInBufferSize, piop->nOutBufferSize, &dwRet);    
			if (err==0) {
				if (piop->nOutBufferSize)
					copy_to_user((void *)(ULONG_PTR)piop->lpOutBuffer, 
							ke_area + piop->nInBufferSize, 
							piop->nOutBufferSize);
				if (piop->lpBytesReturned)
					copy_to_user((void *)(ULONG_PTR)piop->lpBytesReturned, &dwRet, sizeof(DWORD));
			
				kfree(ke_area);
				return length;
			}
			else {
				KdPrintW(("Kernel_ioctl(): return %d\n", err));
			}
			
			kfree(ke_area);
            return -EINVAL;
            
        } else {
            KdPrintW(("Wrong signature: %x\n", piop->Magic));
            return -EINVAL;
        }
#endif
    }
    
    return -EINVAL;
}

typedef struct _hpt_GET_INFO {
	char *buffer;
	int buflength;
	int bufoffset;
	int buffillen;
	int filpos;	
} HPT_GET_INFO;

void hpt_copy_mem_info(HPT_GET_INFO *pInfo, char *data, int datalen)
{
	if (pInfo->filpos < pInfo->bufoffset) {	/* Current offset before buffer offset */
		if (pInfo->filpos + datalen <= pInfo->bufoffset) {
			pInfo->filpos += datalen; 		/* Discard if completely before buffer */
			return;
		} else {				/* Partial copy, set to begin */
			data += (pInfo->bufoffset - pInfo->filpos);
			datalen  -= (pInfo->bufoffset - pInfo->filpos);
			pInfo->filpos = pInfo->bufoffset;
		}
	}
	
	pInfo->filpos += datalen;		/* Update current offset */
	if (pInfo->buffillen == pInfo->buflength) /* Buffer full, discard */
		return;
	
	if (pInfo->buflength - pInfo->buffillen < datalen)  /* Overflows buffer ? */
	datalen = pInfo->buflength - pInfo->buffillen;
	
	memcpy(pInfo->buffer + pInfo->buffillen, data, datalen);
	pInfo->buffillen += datalen;
}

#define shortswap(w) ((WORD)((w)>>8 | ((w) & 0xFF)<<8))

static void get_disk_name(char *name, PDevice pDev)
{
	int i;
	for (i = 0; i < 10; i++) 
		((WORD*)name)[i] = shortswap(pDev->IdentifyData.ModelNumber[i]);
	name[20] = '\0';
}

int hpt_copy_info(HPT_GET_INFO *pinfo, char *fmt, ...) 
{
	va_list args;
	char buf[128];
	int len;
  
	va_start(args, fmt);
	len = vsprintf(buf, fmt, args);
	va_end(args);
	hpt_copy_mem_info(pinfo, buf, len);
	return len;
}

static void hpt_copy_disk_info(HPT_GET_INFO *pinfo, PVDevice pVDev)
{
	UINT iBus, iChip, iChan, iDev;
	char name[32], arrayname[16];
	UINT chan = 0;
	get_disk_name(name, &pVDev->u.disk);
	    arrayname[0]=0;
	
	for(iBus = 0; iBus < MAX_VBUS; iBus++) 			
		for(iChip = 0; iChip < gVBus[iBus].nInstances; iChip++)
			for(iChan = 0; iChan < MAX_CHANNEL_PER_CHIP; iChan++)
	        {
				for(iDev = 0; iDev < MAX_DEVICE_PER_CHANNEL; iDev++)
				    if(&pVDev->u.disk == gVBus[iBus].pChipInstance[iChip]->pChannel[iChan]->pDevice[iDev])
					{
						 hpt_copy_info(pinfo, "%d Channel %d %s %s  %5dMB  %s %s\n",
							  (iChip>>1)+1, (chan+1), 
							  pVDev->u.disk.df_sata? "      " : iDev?"Slave ":"Master", 
							  name, pVDev->VDeviceCapacity>>11, 
							  ((!pVDev->u.disk.df_on_line)? "Disabled" : 
							  ((pVDev->VDeviceType != VD_SPARE)?"Normal  ":"Spare   ")), arrayname);
						 return;
					}
				 chan++;
				 chan &= ~4;
			}
}


static void get_sd_name(char *name, struct Scsi_Host *host, int id)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,0)
	Scsi_Device *SDptr;
	int first_devno = 0;
	name[0] = 0;
	SDptr = host->host_queue;
	if (SDptr && SDptr->device_queue) 
		first_devno = MINOR(SDptr->device_queue->request.rq_dev)/16;
	while (SDptr) {
	   if (SDptr->id == id) {
		   if(SDptr->device_queue) {
			   int devno = (SDptr->device_queue->request.rq_dev)?(MINOR(SDptr->device_queue->request.rq_dev)/16):(first_devno+id);
			   sprintf(name, "/dev/sd%c", 'a' + devno);
		   }
		   break;
	   }
	   SDptr = SDptr->next;
	}
#else 
	struct block_device *bdev = get_bdev(host, id);
	name[0] = 0;
	if (bdev) {
		sprintf(name, "/dev/%s", bdev->bd_disk->disk_name);
		blkdev_put(bdev __BDEV_RAW);
	}
#endif
}

int hpt_get_info(char *buffer,char **start,off_t offset,int length)
{
	HPT_GET_INFO info;
	UINT iBus,iChip,j,i;
	PDevice pDev;
	PVDevice pVDev;
	static char sd[32];
 
	info.buffer     = buffer;
	info.buflength  = length;
	info.bufoffset  = offset;
	info.filpos     = 0;
	info.buffillen  = 0;

	if (start) *start = buffer;

	hpt_copy_info(&info, "%s Version %s\n\n", DRIVER_NAME, DRIVER_VERSION);

	hpt_copy_info(&info, "Physical device list\n");
#ifdef R1540
	hpt_copy_info(&info, "Controller/Bus/ID  Model                Capacity  Status  \n");
	hpt_copy_info(&info, "-----------------------------------------------------------\n");
#else 
	hpt_copy_info(&info, "Controller/Bus/ID  Model                Capacity  Status   Array\n");
	hpt_copy_info(&info, "-------------------------------------------------------------------\n");
#endif
	for(iBus = 0; iBus < MAX_VBUS; iBus++) 			
		for(iChip = 0; iChip < gVBus[iBus].nInstances; iChip++)
			for(j = 0; j < MAX_CHANNEL_PER_CHIP; j++)
				for(i = 0; i < MAX_DEVICE_PER_CHANNEL; i++)
	            {
					pDev=gVBus[iBus].pChipInstance[iChip]->pChannel[j]->pDevice[i];
					if(pDev)
					   hpt_copy_disk_info(&info, Map2pVDevice(pDev));
				}

	hpt_copy_info(&info, "\nLogical device list\n");
	hpt_copy_info(&info, "No. sd         Type         Name                 Capacity  Status\n");
	hpt_copy_info(&info, "-------------------------------------------------------------------\n");
	
	j=1;
	for(iBus = 0; iBus < MAX_VBUS; iBus++) 
		for(i = 0; i < MAX_VDEVICE_PER_VBUS; i++){
            pVDev = gVBus[iBus].pVDevice[i];
			if(pVDev){
				ZeroMemory(sd,32);
				j=i+1;
				get_sd_name(sd, hpt_vhost, i);
				{
					char name[32];
					get_disk_name(name, &pVDev->u.disk);
					hpt_copy_info(&info, 
						"%2d  %3s   %s  %s  %5dMB  %s\n",
						j, sd, "Single disk", name, pVDev->VDeviceCapacity>>11, 
						/* gmm 2001-6-19: Check if pVDev has been added to an array. */
						((pVDev->pParent) ? "Unavailable" : "Normal"));
				 }
	         }
	     }
	

	
	return info.buffillen;
}

/*******************************************************************************
 * hpt3xx_ProcInfo
 * Parameters:
 *  *buffer: I/O buffer
 *  **start: if inout == FALSE pointer into buffer where user read should start
 *  offset: current offset into a /proc/scsi/hpt37x2/[0...] file
 *  length: length of buffer
 *  hostno: Scsi_Host host_no
 *  inout: TRUE - user is writing; FALSE - user is reading
 * Returns:
 *   Length of data
 *****/
int hpt3xx_ProcInfo(char *buffer,char **start,off_t offset,int length,   
                   int hostno,int inout)
{
	KdPrintI(("hpt3xx_ProcInfo() length %d ha %d offs %d inout %d\n",
		length,hostno,(int)offset,inout));

	if (inout)
		return(hpt_set_info(buffer,length));
	else
		return(hpt_get_info(buffer,start,offset,length));
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,5,0)
int hpt3xx_ProcInfo26(struct Scsi_Host *host, char *buffer, char **start, 
			off_t offset, int length, int inout)
{
	return hpt3xx_ProcInfo(buffer, start, offset, length, host->host_no, inout);
}
#endif
