/***************************************************************************
* File:          	ioctl.c
* Description:   	ioctl interface implementation
* Dependence:    	hptintf.h
***************************************************************************/
#include "gui_lib.c"
#include "oslinux.h"

extern struct semaphore  hpt_queue_wt;
extern atomic_t DPC_Request_Nums;

/*--------------------------------------------------------------------------------------------------------- */
#pragma pack(1)

typedef struct _HPT_REBUILD_PARAM
{
	DEVICEID idMirror;
	DWORD Lba;
	UCHAR nSector;
} HPT_REBUILD_PARAM, *PHPT_REBUILD_PARAM;

#pragma pack()

#define MAX_EVENTS 10
static HPT_EVENT hpt_event_queue[MAX_EVENTS];
static int event_queue_head=0, event_queue_tail=0;

/*---------------------------------------------------------------------------------------------------------- */
static int event_queue_add(PHPT_EVENT pEvent)
{
	int p;

	p = (event_queue_tail + 1) % MAX_EVENTS;
	if (p==event_queue_head)
	{
		return -1;
	}
	hpt_event_queue[event_queue_tail] = *pEvent;
	event_queue_tail = p;
	return 0;
}

static int event_queue_remove(PHPT_EVENT pEvent)
{
	if (event_queue_head != event_queue_tail)
	{
		*pEvent = hpt_event_queue[event_queue_head];
		event_queue_head++;
		event_queue_head %= MAX_EVENTS;
		return 0;
	}
	return -1;
}

void HPTLIBAPI ioctl_ReportEvent(UCHAR event, PVOID param)
{
	HPT_EVENT e;
	ZeroMemory(&e, sizeof(e));
	e.EventType = event;
	switch(event)
	{
		case ET_DEVICE_REMOVED:
		case ET_DEVICE_PLUGGED:
			e.DeviceID = VDEV_TO_ID((PVDevice)param);
			break;

		default:
			break;
	}
	event_queue_add(&e);
}

static int hpt_device_io(DEVICEID id, int cmd, ULONG lba, DWORD nSector, PVOID buffer);
static int hpt_get_event(PHPT_EVENT pEvent);
int hpt_set_array_state(DEVICEID idArray, DWORD state);
int hpt_rescan_all(void);

int Kernel_DeviceIoControl(
							DWORD dwIoControlCode,       	/* operation control code */
							PVOID lpInBuffer,            	/* input data buffer */
							DWORD nInBufferSize,         	/* size of input data buffer */
							PVOID lpOutBuffer,           	/* output data buffer */
							DWORD nOutBufferSize,        	/* size of output data buffer */
							PDWORD lpBytesReturned      	/* byte count */
						)
{
	switch(dwIoControlCode)
	{
		case HPT_IOCTL_DEVICE_IO: /*reserve? */
		{
			PVOID buffer;
			DEVICEID id;
			int cmd;
			ULONG lba;
			DWORD nSector;
			int iSuccess;

			if (nInBufferSize < sizeof(HPT_DEVICE_IO)) return -1;

			id		=	((PHPT_DEVICE_IO)lpInBuffer)->id;
			cmd		=	((PHPT_DEVICE_IO)lpInBuffer)->cmd;
			lba		=	((PHPT_DEVICE_IO)lpInBuffer)->lba;
			nSector	=	((PHPT_DEVICE_IO)lpInBuffer)->nSector;
			buffer  =	((PHPT_DEVICE_IO)lpInBuffer)->buffer;

		    iSuccess = hpt_device_io(id, cmd, lba, nSector, buffer);

			if(cmd == IO_COMMAND_READ)
			{
				memcpy(lpOutBuffer, &iSuccess, sizeof(int));
				memcpy(lpOutBuffer+4, buffer, nSector*512);
			}
			else
			{
				*(int*)lpOutBuffer = iSuccess;
			}

			if(iSuccess != 0)
				return -1;
		}
		break;

		case HPT_IOCTL_GET_EVENT:
		{
			PHPT_EVENT pInfo;

			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=sizeof(HPT_EVENT)) return -1;

			pInfo = (PHPT_EVENT)lpOutBuffer;

			if (hpt_get_event(pInfo)!=0)
				return -1;
		}
		break;
		case HPT_IOCTL_RESCAN_DEVICES:
		{
			if (nInBufferSize!=0) return -1;
			if (nOutBufferSize!=0) return -1;
			hpt_rescan_all();
		}
		break;

		case HPT_IOCTL_LOCK_DEVICE:
			return -1;
	
		case HPT_IOCTL_UNLOCK_DEVICE:
			return -1;

		default:
		{
			if(hpt_default_ioctl(dwIoControlCode, lpInBuffer, nInBufferSize, 
				lpOutBuffer, nOutBufferSize, lpBytesReturned) == -1) return -1;
            return 0;
		}
	}

	if (lpBytesReturned)
		*lpBytesReturned = nOutBufferSize;
	return 0;
}

/*---------------------------------------------------------------------------------------------------------- */
int hpt_device_io(DEVICEID id, int cmd, ULONG lba, DWORD nSector, PVOID buffer)
{
	return -1;
}

int hpt_get_event(PHPT_EVENT pEvent)
{
	return event_queue_remove(pEvent);
}


void fRescanAllDevice(_VBUS_ARG int quick_probe);
void Check_Idle_Call(void);

void HPTLIBAPI do_rescan(_VBUS_ARG void *unused)
{
	fRescanAllDevice(0);
}

int hpt_rescan_all()
{
    unsigned long flags;
	unsigned long timeout;
    
    spin_lock_irqsave(&driver_lock, flags);
    CallWhenIdle(_VBUS_P do_rescan, 0);
    spin_unlock_irqrestore(&driver_lock, flags);

	/* wait the rescan to complete? */
	timeout = jiffies + 60 * HZ;
    while((CURRCMDs || atomic_read(&DPC_Request_Nums)) && time_before(jiffies, timeout)) {
    	schedule();
    	if (!mWaitingForIdle(gVBus)) return 0;
    }

	spin_lock_irqsave(&driver_lock, flags);
	if((CURRCMDs == 0) && (atomic_read(&DPC_Request_Nums) == 0))
	    Check_Idle_Call();
	spin_unlock_irqrestore(&driver_lock, flags);

    return 0;
}

/***************************************************************************
 * Function:     hpt_rebuild_data_block
 * Description:  rebuild a mirror block
 * Returns:      None
 ***************************************************************************/

